package net.rumati.logging.muffero;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Date;

/**
 * Represents a single log statement, and the application state around that
 * logging statement. {@code LogEvent}s are immutable. All setters that would
 * normally change object state return a new instance with the changes applied.
 * @author Jesse Long
 */
public class LogEvent
    implements Serializable
{
    private static final long serialVersionUID = 2l;
    
    /**
     * The name of the logger which was used to log this message.
     */
    private final String loggerName;
    
    /**
     * The priority of the logging statement
     */
    private final Priority priority;
    
    /**
     * The text message logged.
     */
    private final String message;
    
    /**
     * The {@link java.lang.Throwable} logged, if any.
     */
    private final Throwable throwable;
    
    /**
     * The time that the {@link LogEvent} was created, in milliseconds from Unix epoch.
     */
    private final long time;
    
    /**
     * The name of the {@link Thread} on which the logging statement was made.
     */
    private final String threadName;
    
    /**
     * The id of the {@link Thread} on which the logging statement was made.
     */
    private final long threadId;
    
    /**
     * The stack trace of the method that performed the logging statement.
     */
    private final StackTraceElement[] stackTrace;

    /**
     * Private constructor used in setters to enforce immutability.
     * @param loggerName
     * @param priority
     * @param message
     * @param throwable
     * @param time
     * @param threadName
     * @param threadId
     * @param stackTrace 
     */
    private LogEvent(String loggerName, Priority priority, String message, Throwable throwable, long time, String threadName,
                    long threadId, StackTraceElement[] stackTrace)
    {
        this.loggerName = loggerName;
        this.priority = priority;
        this.message = message;
        this.throwable = throwable;
        this.time = time;
        this.threadName = threadName;
        this.threadId = threadId;
        this.stackTrace = stackTrace;
    }

    /**
     * Package private constructor, because you dont need to create custom instances.
     * @param fqcn
     * @param loggerName
     * @param priority
     * @param message
     * @param throwable 
     */
    LogEvent(String fqcn, String loggerName, Priority priority, String message, Throwable throwable)
    {
        this.loggerName = loggerName;
        this.priority = priority;
        this.message = message;
        this.throwable = throwable;
        this.time = new Date().getTime();
        this.threadName = Thread.currentThread().getName();
        this.threadId = Thread.currentThread().getId();

        StackTraceElement[] st = Thread.currentThread().getStackTrace();
        int idx = getIndexOfCaller(fqcn, st);
        if (idx < 0){
            this.stackTrace = st;
        }else{
            this.stackTrace = new StackTraceElement[st.length - idx];
            for (int i = 0; i < this.stackTrace.length; i++){
                this.stackTrace[i] = st[i + idx];
            }
        }
    }

    private static int getIndexOfCaller(String fqcn, StackTraceElement[] st)
    {
        for (int i = 0; i < st.length; i++){
            if (st[i].getClassName().equals(fqcn)){
                for (; i < st.length; i++){
                    if (!st[i].getClassName().equals(fqcn)){
                        return i;
                    }
                }
                return -1;
            }
        }
        return -1;
    }

    /**
     * Returns the name of the {@link Logger} that created this {@link LogEvent}.
     * @return the name of the {@link Logger} that created this {@link LogEvent}.
     */
    public String getLoggerName()
    {
        return loggerName;
    }

    /**
     * Creates a new {@link LogEvent} instance, with the provided {@link Logger} name
     * @param loggerName the name of the {@link Logger} used to create this {@link LogEvent}.
     * @return a new {@link LogEvent} instance, with the provided {@link Logger} name
     */
    public LogEvent setLoggerName(String loggerName)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, threadName, threadId, stackTrace);
    }

    public Priority getPriority()
    {
        return priority;
    }

    public LogEvent setPriority(Priority priority)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, threadName, threadId, stackTrace);
    }

    public String getMessage()
    {
        return message;
    }

    public LogEvent setMessage(String message)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, threadName, threadId, stackTrace);
    }

    public Throwable getThrowable()
    {
        return throwable;
    }

    public LogEvent setThrowable(Throwable throwable)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, threadName, threadId, stackTrace);
    }

    public long getTime()
    {
        return time;
    }

    public LogEvent setTime(long time)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, threadName, threadId, stackTrace);
    }

    public Date getDate()
    {
        return new Date(time);
    }

    public LogEvent setDate(Date date)
    {
        return new LogEvent(loggerName, priority, message, throwable, date.getTime(), threadName, threadId, stackTrace);
    }

    public String getThreadName()
    {
        return threadName;
    }
    
    public long getThreadId()
    {
        return threadId;
    }

    public LogEvent setThread(Thread thread)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, thread.getName(), thread.getId(), stackTrace);
    }

    public StackTraceElement[] getStackTrace()
    {
        return stackTrace.clone();
    }

    /**
     * Sets stack trace of the method that called the {@link Logger}.
     * @param stackTrace The new stack trace
     * @return A new {@link LogEvent} instance which is equivalent 
     */
    public LogEvent setStackTrace(StackTraceElement[] stackTrace)
    {
        return new LogEvent(loggerName, priority, message, throwable, time, threadName, threadId, stackTrace.clone());
    }

    @Override
    public boolean equals(Object obj)
    {
        if (obj == null){
            return false;
        }
        if (getClass() != obj.getClass()){
            return false;
        }
        final LogEvent other = (LogEvent)obj;
        if ((this.loggerName == null) ? (other.loggerName != null) : !this.loggerName.equals(other.loggerName)){
            return false;
        }
        if (this.priority != other.priority){
            return false;
        }
        if ((this.message == null) ? (other.message != null) : !this.message.equals(other.message)){
            return false;
        }
        if (this.throwable != other.throwable && (this.throwable == null || !this.throwable.equals(other.throwable))){
            return false;
        }
        if (this.time != other.time){
            return false;
        }
        if ((this.threadName == null) ? (other.threadName != null) : !this.threadName.equals(other.threadName)){
            return false;
        }
        if (this.threadId != other.threadId){
            return false;
        }
        if (!Arrays.deepEquals(this.stackTrace, other.stackTrace)){
            return false;
        }
        return true;
    }

    @Override
    public int hashCode()
    {
        int hash = 3;
        hash = 47 * hash + (this.loggerName != null ? this.loggerName.hashCode() : 0);
        hash = 47 * hash + (this.priority != null ? this.priority.hashCode() : 0);
        hash = 47 * hash + (this.message != null ? this.message.hashCode() : 0);
        hash = 47 * hash + (this.throwable != null ? this.throwable.hashCode() : 0);
        hash = 47 * hash + (int)(this.time ^ (this.time >>> 32));
        hash = 47 * hash + (this.threadName != null ? this.threadName.hashCode() : 0);
        hash = 47 * hash + (int)(this.threadId ^ (this.threadId >>> 32));
        hash = 47 * hash + Arrays.deepHashCode(this.stackTrace);
        return hash;
    }
    
    
}
