package net.rumati.logging.muffero;

import java.util.List;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import net.rumati.logging.muffero.appender.Appender;
import net.rumati.logging.muffero.appender.CompoundAppender;

public class Logger
{
    private static final String FQCN = Logger.class.getName();
    private final String name;
    private final ReadWriteLock lock = new ReentrantReadWriteLock(true);
    private boolean inherit = true;
    private Logger parent;
    private final CompoundAppender appenders = new CompoundAppender();

    Logger(Logger parent, String name)
    {
        this.parent = parent;
        this.name = name;
    }

    public String getName()
    {
        return name;
    }

    public void setInheritParentAppenders(boolean inherit)
    {
        lock.writeLock().lock();
        try{
            this.inherit = inherit;
        }finally{
            lock.writeLock().unlock();
        }
    }

    public boolean inheritsParentAppenders()
    {
        lock.readLock().lock();
        try{
            return inherit;
        }finally{
            lock.readLock().unlock();
        }
    }

    public void addAppender(Appender appender)
    {
        appenders.addAppender(appender);
    }

    public void removeAppender(Appender appender)
    {
        appenders.removeAppender(appender);
        appender.shutdown();
    }

    /**
     * Returns an unmodifiable list of {@link Appender}s associated with this {@link Logger}. The returned
     * {@link List} is unmodifiable. To add and remove {@link Appender}s, use the
     * {@link #addAppender(net.rumati.logging.muffero.appender.Appender)} and 
     * {@link #removeAppender(net.rumati.logging.muffero.appender.Appender)} methods.
     * @return An unmodifiable list of appenders associated with this {@link Logger}.
     */
    public List<Appender> getAppenders()
    {
        return appenders.getAppenders();
    }

    /**
     * Returns most immediate existing ancestor. For example, if this logger is
     * <code>a.b.c</code>, and logger <code>a.b</code> exists, then logger <code>a.b</code>
     * will be returned. If logger <code>a.b</code> does not exist, but logger
     * <code>a</code> does exist, them logger <code>a</code> will be returned. If
     * neither <code>a.b</code> nor <code>a</code> exist, then the root logger is
     * returned.
     * @return The most immediate existing ancestor
     */
    public Logger getParent()
    {
        lock.readLock().lock();
        try{
            return parent;
        }finally{
            lock.readLock().unlock();
        }
    }

    void setParent(Logger parent)
    {
        lock.writeLock().lock();
        try{
            this.parent = parent;
        }finally{
            lock.writeLock().unlock();
        }
    }

    private boolean isPriorityEnabled(Priority p)
    {
        if (appenders.isPriorityEnabled(name, p)){
            return true;
        }
        if (inheritsParentAppenders()){
            return getParent().isPriorityEnabled(p);
        }else{
            return false;
        }
    }

    public boolean isTraceEnabled()
    {
        return isPriorityEnabled(Priority.TRACE);
    }

    public void trace(String msg)
    {
        trace(msg, null);
    }

    public void trace(String msg, Throwable t)
    {
        log(Priority.TRACE, FQCN, msg, t);
    }

    public boolean isDebugEnabled()
    {
        return isPriorityEnabled(Priority.DEBUG);
    }

    public void debug(String msg)
    {
        debug(msg, null);
    }

    public void debug(String msg, Throwable t)
    {
        log(Priority.DEBUG, FQCN, msg, t);
    }

    public boolean isInfoEnabled()
    {
        return isPriorityEnabled(Priority.INFO);
    }

    public void info(String msg)
    {
        info(msg, null);
    }

    public void info(String msg, Throwable t)
    {
        log(Priority.INFO, FQCN, msg, t);
    }

    public boolean isWarnEnabled()
    {
        return isPriorityEnabled(Priority.WARNING);
    }

    public void warn(String msg)
    {
        warn(msg, null);
    }

    public void warn(String msg, Throwable t)
    {
        log(Priority.WARNING, FQCN, msg, t);
    }

    public boolean isErrorEnabled()
    {
        return isPriorityEnabled(Priority.ERROR);
    }

    public void error(String msg)
    {
        error(msg, null);
    }

    public void error(String msg, Throwable t)
    {
        log(Priority.ERROR, FQCN, msg, t);
    }

    public void log(Priority priority, String fqcn, String message, Throwable t)
    {
        if (priority.isMoreCriticalThan(Priority.ERROR)){
            throw new IllegalArgumentException("Cannot log at level: " + priority.name());
        }

        LogEvent event = new LogEvent(fqcn, name, priority, message, t);

        log(event);
    }

    private void log(LogEvent event)
    {
        appenders.append(event);

        if (inheritsParentAppenders()){
            getParent().log(event);
        }
    }

    @Override
    public String toString()
    {
        return "Logger: " + name;
    }
}
