package net.rumati.logging.muffero.config.prop;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import net.rumati.logging.muffero.LoggingException;
import net.rumati.logging.muffero.appender.Appender;
import net.rumati.logging.muffero.layout.Layout;

public final class PropertyConfigContext
{
    private final List<PropertyObjectLoader> loaders;
    private final Map<String, String> rootConfig;
    private final List<NamedObjectCache<?>> namedObjectCaches = new LinkedList<NamedObjectCache<?>>();

    PropertyConfigContext(List<PropertyObjectLoader> loaders, Map<String, String> config)
    {
        this.loaders = loaders;
        this.rootConfig = config;
        
        registerNamedObjectCache(Appender.class, "muffero.appender");
        registerNamedObjectCache(Layout.class, "muffero.layout");
    }
    
    public <T> void registerNamedObjectCache(Class<T> type, String configPrefix)
    {
        namedObjectCaches.add(new NamedObjectCache<T>(configPrefix, type));
    }
    
    @SuppressWarnings("unchecked")
    private <T> NamedObjectCache<T> getNamedObjectCache(Class<T> type)
    {
        for (NamedObjectCache<?> c : namedObjectCaches){
            if (c.getType() == type){
                return (NamedObjectCache<T>)c;
            }
        }
        
        return null;
    }

    private <T> T loadNameableObject(Class<T> objectType, String name, String configPrefix, Map<String, String> config)
            throws LoggingException
    {
        NamedObjectCache<T> cache = getNamedObjectCache(objectType);
        
        if (cache != null){
            T retv = cache.loadNamedObject(name);
            
            if (retv != null){
                return retv;
            }
        }
        
        return loadObjectWithConfigPrefix(name, objectType, configPrefix, config);
    }

    public Appender loadAppender(String name, String configPrefix, Map<String, String> config)
            throws LoggingException
    {
        return loadNameableObject(Appender.class, name, configPrefix, config);
    }

    public Layout loadLayout(String name, String configPrefix, Map<String, String> config)
            throws LoggingException
    {
        return loadNameableObject(Layout.class, name, configPrefix, config);
    }

    public <T> T loadObjectWithConfigPrefix(String className, Class<T> typeClass, String configPrefix, Map<String, String> config)
            throws LoggingException
    {
        Map<String, String> resultConfig = new TreeMap<String, String>();
        for (Map.Entry<String, String> e : config.entrySet()){
            if (e.getKey().startsWith(configPrefix)){
                resultConfig.put(e.getKey().substring(configPrefix.length()), e.getValue());
            }
        }

        for (PropertyObjectLoader loader : loaders){
            if (loader.canLoad(className, typeClass)){
                return loader.loadObject(className, typeClass, resultConfig, this);
            }
        }

        try{
            Class c = Class.forName(className);
            if (!typeClass.isAssignableFrom(c)){
                throw new LoggingException("Requested class \"" + className + "\" does not implement \"" + typeClass.getName() + "\"");
            }
            @SuppressWarnings("unchecked")
            Class<? extends T> ctc = c;
            try{
                return ctc.newInstance();
            }catch (IllegalAccessException e){
                throw new LoggingException("Error instantiating object " + className, e);
            }catch (InstantiationException e){
                throw new LoggingException("Error instantiating object " + className, e);
            }catch (ExceptionInInitializerError e){
                throw new LoggingException("Error instantiating object " + className, e);
            }
        }catch (ClassNotFoundException e){
            throw new LoggingException(e);
        }
    }

    private class NamedObjectCache<T>
    {
        private final String namedConfigPrefix;
        private final Class<T> type;
        private final Map<String, T> cache = new TreeMap<String, T>();

        public NamedObjectCache(String namedConfigPrefix, Class<T> type)
        {
            this.namedConfigPrefix = namedConfigPrefix;
            this.type = type;
        }

        public Class<T> getType()
        {
            return type;
        }

        public T loadNamedObject(String name) throws LoggingException
        {
            if (name.contains(".")){
                return null;
            }

            if (cache.containsKey(name)){
                return cache.get(name);
            }

            String configKey = namedConfigPrefix + "." + name;

            String className = rootConfig.get(configKey);
            if (className == null){
                return null;
            }
            
            T retv = loadObjectWithConfigPrefix(className, type, configKey + ".", rootConfig);

            cache.put(name, retv);

            return retv;
        }
    }
}
